#include <Wire.h>
#include <Adafruit_GFX.h>
#include "Adafruit_LEDBackpack.h"
#include <SPI.h>
#include <RH_RF69.h>
#include <RHReliableDatagram.h>
#include <Adafruit_NeoPixel.h>

// === NeoPixel Setup ===
#define LED_PIN    PIN_NEOPIXEL
#define NUM_LEDS   1
Adafruit_NeoPixel strip(NUM_LEDS, LED_PIN, NEO_GRB + NEO_KHZ800);
static const uint32_t COLOR_OFF = 0x000000;         // Off
static const uint32_t COLOR_MY_TURN    = 0x003200;  // Green
static const uint32_t COLOR_ENEMY_TURN = 0x320000;  // Red
static const uint32_t COLOR_GAME_OVER  = 0x000032;  // Blue 

bool isIdle = false;
unsigned long previousMillisIdle = 0;
const unsigned long intervalIdle = 40000;

// === Basic Setup ===
// Color definitions
#define LED_BLUE 1
#define LED_PINK 2
#define LED_ORANGE 3
#define BTN_UP A6
#define BTN_DOWN A0
#define BTN_DRAW A2
#define BTN_LEFT A1
#define BTN_RIGHT A3

// === Radio Setup ===
// Change this for each radio - one radio's MY_ADDRESS should be 1, OTHER_ADDRESS 2, the other's MY_ADDRESS should be 2, OTHER_ADDRESS 1
#define MY_ADDRESS      1  
#define OTHER_ADDRESS   2  
#define RFM69_CS PIN_SPI1_MISO
#define RFM69_IRQ A7
#define RFM69_RST PIN_SPI1_SCK
RH_RF69 rf69(RFM69_CS, RFM69_IRQ);
RHReliableDatagram manager(rf69, MY_ADDRESS);
#define MAX_RETRIES 3
uint8_t buf[RH_RF69_MAX_MESSAGE_LEN];
uint8_t replyBuf[RH_RF69_MAX_MESSAGE_LEN];


// Game board size
#define BOARD_WIDTH 8
#define BOARD_HEIGHT 4
// Define message types
#define MSG_ATTACK     'A'    // Attack message
#define MSG_HIT        'H'    // Hit response
#define MSG_MISS       'M'    // Miss response
#define MSG_GAMEOVER   'G'    // Game over message
#define MSG_RESTART    'R'    // Restart game message
char myShipsBoard[BOARD_HEIGHT][BOARD_WIDTH] = {0};     // My ships: 'S'=ship, 'H'=hit, 'M'=miss
char myAttackBoard[BOARD_HEIGHT][BOARD_WIDTH] = {0};     // My ships: 'H'=hit, 'M'=miss

// Game state
bool myTurn;
bool gameOver = false;
int myShipCells = 0;  // Total cells occupied by ships
int myHitCells = 0;   // Total hits on my ships
int enemyHitCells = 0; // Total hits on enemy ships

void drawMyAttackBoard()
{
  for (int y = 0; y < BOARD_HEIGHT; y++) {
    for (int x = 0; x < BOARD_WIDTH; x++) {
      char cell = myAttackBoard[y][x];
      if (cell == 'H') {drawPixelRightScreen(x,y,LED_ORANGE);} 
      if (cell == 'M') {drawPixelRightScreen(x,y,LED_PINK);} 
    }
  }
}

#define NUM_SHIPS 2
int shipLengths[NUM_SHIPS] = {2, 3}; // Destroyer, Cruiser

void clearMyBoard()
{
  for (int y = 0; y < BOARD_HEIGHT; y++) {
    for (int x = 0; x < BOARD_WIDTH; x++) {
      myShipsBoard[y][x] = 0;
      myAttackBoard[y][x] = 0;
    }
  }
  myShipCells = 0;
  myHitCells = 0;
  enemyHitCells = 0;
}

// Check if a ship can be placed at specified position and orientation
bool canPlaceShip(int startX, int startY, int length, bool horizontal) {
  if (horizontal) {
    if (startX + length > BOARD_WIDTH) return false;
    
    for (int i = 0; i < length; i++) {
      if (myShipsBoard[startY][startX + i] == 'S') return false;
    }
  } else {
    if (startY + length > BOARD_HEIGHT) return false;
    
    for (int i = 0; i < length; i++) {
      if (myShipsBoard[startY + i][startX] == 'S') return false;
    }
  }
  return true;
}

void myShipBoardInit()
{
  clearMyBoard();
  
  // Place ships randomly
  for (int i = 0; i < NUM_SHIPS; i++) {
    int length = shipLengths[i];
    bool placed = false;
    
    while (!placed) {
      bool horizontal = random(2) == 0;  // 0=horizontal, 1=vertical
      int startX, startY;
      
      if (horizontal) {
        startX = random(BOARD_WIDTH - length + 1);
        startY = random(BOARD_HEIGHT);
      } else {
        startX = random(BOARD_WIDTH);
        startY = random(BOARD_HEIGHT - length + 1);
      }
      
      if (canPlaceShip(startX, startY, length, horizontal)) {
        // Place the ship
        if (horizontal) {
          for (int j = 0; j < length; j++) {
            myShipsBoard[startY][startX + j] = 'S';
          }
        } else {
          for (int j = 0; j < length; j++) {
            myShipsBoard[startY + j][startX] = 'S';
          }
        }
        placed = true;
        myShipCells += length;
      }
    }
  }
}

// Send attack coordinates
void sendAttack(int x, int y) {
  if (gameOver) return;
  
  uint8_t data[4];
  data[0] = MSG_ATTACK;
  data[1] = x;
  data[2] = y;
  data[3] = '\0';
  
  if (manager.sendtoWait(data, 4, OTHER_ADDRESS)) {
    Serial.println("Attack sent successfully");
    setMyTurnStatus(false);
  } else {
    Serial.println("Attack send failed");
    setMyTurnStatus(true); // Try again
  }
}

// Send hit response
void sendHitResponse(int x, int y) {
  uint8_t data[4];
  data[0] = MSG_HIT;
  data[1] = x;
  data[2] = y;
  data[3] = '\0';
  
  if (manager.sendtoWait(data, 4, OTHER_ADDRESS)) {
    Serial.println("Hit response sent");
  } else {
    Serial.println("Hit response failed");
  }
}

// Send miss response
void sendMissResponse(int x, int y) {
  uint8_t data[4];
  data[0] = MSG_MISS;
  data[1] = x;
  data[2] = y;
  data[3] = '\0';
  
  if (manager.sendtoWait(data, 4, OTHER_ADDRESS)) {
    Serial.println("Miss response sent");
  } else {
    Serial.println("Miss response failed");
  }
}

// Send game over message
void sendGameOver() {
  uint8_t data[2];
  data[0] = MSG_GAMEOVER;
  data[1] = '\0';
  
  if (manager.sendtoWait(data, 2, OTHER_ADDRESS)) {
    Serial.println("Game over message sent");
  } else {
    Serial.println("Game over message failed");
  }
}

// Send restart game message
void sendRestartGame() {
  uint8_t data[2];
  data[0] = MSG_RESTART;
  data[1] = '\0';
  
  if (manager.sendtoWait(data, 2, OTHER_ADDRESS)) {
    Serial.println("Restart game message sent");
    restartGame();
  } else {
    Serial.println("Restart game message failed");
  }
}

// Process attack at coordinates
void processAttack(int x, int y) {
  Serial.print("Enemy attacked (");
  Serial.print(x);
  Serial.print(",");
  Serial.print(y);
  Serial.println(")");
  
  if (myShipsBoard[y][x] == 'S') {
    // Hit
    myShipsBoard[y][x] = 'H';
    myHitCells++;
    Serial.println("Hit!");
    sendHitResponse(x, y);
    
    // Check if all ships are sunk
    // i dont think i need this
    // if (myHitCells >= myShipCells) {
    //   Serial.println("All my ships are sunk! Game over!");
    //   setGameOver(false); // I lost
    // }
  } else {
    // Miss
    myShipsBoard[y][x] = 'M';
    Serial.println("Miss!");
    sendMissResponse(x, y);
  }
  
  redrawScreen();
  if (!gameOver) {
    setMyTurnStatus(true);
  }
}

// Restart the game
void restartGame() {
  gameOver = false;
  clearMyBoard();
  myShipBoardInit();
  
  // Determine starting player based on address
  if (MY_ADDRESS == 2) {
    setMyTurnStatus(true);
  } else {
    setMyTurnStatus(false);
  }
  
  redrawScreen();
  Serial.println("Game restarted!");
}

// === Display Setup ===
Adafruit_BicolorMatrix matrix;

// Receive and handle messages
void receiveMessage() {
  uint8_t len = sizeof(buf);
  uint8_t from;
  // isIdle = false;
  
  if (manager.recvfromAckTimeout(buf, &len, 100, &from)) {
    if (from == OTHER_ADDRESS) {
      char msgType = buf[0];
      
      switch (msgType) {
        case MSG_ATTACK: {
          int x = buf[1];
          int y = buf[2];
          processAttack(x, y);
          if (isIdle){setIdle(false);}
          break;
        }
        
        case MSG_HIT: {
          int x = buf[1];
          int y = buf[2];
          myAttackBoard[y][x] = 'H';
          enemyHitCells++;
          Serial.print("Hit at (");
          Serial.print(x);
          Serial.print(",");
          Serial.print(y);
          Serial.println(")!");
          
          // Check if we've sunk all enemy ships
          if (enemyHitCells >= 5) { // 2 + 3 = 5 total ship cells
            Serial.println("All enemy ships sunk! You win!");
            setGameOver(true); // I won
          }
          
          drawMyAttackBoard();
          if (isIdle){setIdle(false);}
          break;
        }
        
        case MSG_MISS: {
          int x = buf[1];
          int y = buf[2];
          myAttackBoard[y][x] = 'M';
          Serial.print("Miss at (");
          Serial.print(x);
          Serial.print(",");
          Serial.print(y);
          Serial.println(").");
          drawMyAttackBoard();
          if (isIdle){setIdle(false);}
          break;
        }
        
        case MSG_GAMEOVER: {
          Serial.println("Game over! You lost all ships!");
          setGameOver(false); // I lost
          if (isIdle){setIdle(false);}
          break;
        }
        
        case MSG_RESTART: {
          Serial.println("Other player restarted the game!");
          restartGame();
          if (isIdle){setIdle(false);}
          break;
        }
      }
    }
  }
}

void drawPixelRightScreen(int x, int y, uint16_t color) {
  /*
    x = 0; y = 0; is bottom left corner
    x range is 0<>7, max x is 7
    y range is 0<>3, max y is 3
  */
  matrix.drawPixel(x, 7-y, color);
}

void drawPixelLeftScreen(int x, int y, uint16_t color) {
  /*
    x = 0; y = 0; is bottom left corner
    x range is 0<>7, max x is 7
    y range is 0<>3, max y is 3
  */
  matrix.drawPixel(7-x, y, color);
}

// Cursor position
int cursorX = 0;
int cursorY = 0;
bool drawnPixels[8][4];
void drawCursor() {
  drawPixelRightScreen(cursorX, cursorY, LED_BLUE);
}
// Button debounce variables
unsigned long lastButtonPress = 0;
unsigned long debounceDelay = 200;

void clearRightScreen()
{
  for (int x = 0; x < 8; x++) {
    for (int y = 0; y < 4; y++) {
      drawnPixels[x][y] = false;
    }
  }
}

void update() {
  if (millis() - lastButtonPress > debounceDelay) {
    bool buttonPressed = false;
    if (isIdle){redrawScreen();};
    if (digitalRead(BTN_UP) == LOW) {
      if (cursorY < 3) {
        cursorY++;
      }
      buttonPressed = true;
    }
    else if (digitalRead(BTN_DOWN) == LOW) {
      if (cursorY > 0) {
        cursorY--;
      }
      buttonPressed = true;
    }
    else if (digitalRead(BTN_LEFT) == LOW) {
      if (cursorX > 0) {
        cursorX--;
      }
      buttonPressed = true;
    }
    else if (digitalRead(BTN_RIGHT) == LOW) {
      if (cursorX < 7) {
        cursorX++;
      }
      buttonPressed = true;
    }
    else if (digitalRead(BTN_DRAW) == LOW) {
      buttonPressed = true;
      
      if (gameOver) {
        // If game is over, restart the game
        sendRestartGame();
      } else if (myTurn) {
        // Send attack if it's my turn
        sendAttack(cursorX, cursorY);
      }
    }
    if (buttonPressed) {
      lastButtonPress = millis();
      previousMillisIdle = millis();
      if (isIdle){setIdle(false);}
      // isIdle = false;
      redrawScreen();
    }
  }
}

int drawWaveX = 0;
int drawColor = 0;
int colors[3] = {LED_GREEN, LED_RED, LED_YELLOW};

void redrawScreen() {
  if(isIdle){
    drawWaveX -= 1;
    if (drawWaveX == -37) {
      drawWaveX = 7;
      drawColor = (drawColor + 1) % 3;
      matrix.setTextColor(colors[drawColor]);
    }
    matrix.clear();
    matrix.setCursor(drawWaveX, 0);
    matrix.print("Hello");
    matrix.writeDisplay();

    return;
  }
  matrix.clear();
  drawMyAttackBoard();
  drawMyShipBoard();
  drawCursor();
  matrix.writeDisplay();
}

void drawMyShipBoard()
{
  for (int y = 0; y < BOARD_HEIGHT; y++) {
    for (int x = 0; x < BOARD_WIDTH; x++) {
      char cell = myShipsBoard[y][x];
      if (cell == 'S') {drawPixelLeftScreen(x,y,LED_BLUE);} 
      if (cell == 'H') {drawPixelLeftScreen(x,y,LED_ORANGE);} 
      if (cell == 'M') {drawPixelLeftScreen(x,y,LED_PINK);} 
    }
  }
}

void setMyTurnStatus(bool status)
{
  if (gameOver) return;
  
  if (status)
  {
    myTurn = true;
    strip.setPixelColor(0, COLOR_MY_TURN);
    strip.show();
  }
  else
  {
    myTurn = false;
    strip.setPixelColor(0, COLOR_ENEMY_TURN);
    strip.show();
  }
}

// Set game over state
void setGameOver(bool iWon) {
  gameOver = true;
  strip.setPixelColor(0, COLOR_GAME_OVER);
  strip.show();
  matrix.clear();
  // Flash victory/defeat pattern
  // for (int i = 0; i < 3; i++) {
  //   matrix.clear();
  if (iWon) {
    // Draw a victory pattern (e.g., "W" for win)
    drawPixelRightScreen(1+1, 2, LED_BLUE);
    drawPixelRightScreen(1+1, 1, LED_BLUE);
    drawPixelRightScreen(3+1, 1, LED_BLUE);
    drawPixelRightScreen(5+1, 1, LED_BLUE);
    drawPixelRightScreen(5+1, 2, LED_BLUE);
    drawPixelRightScreen(2+1, 0, LED_BLUE);
    drawPixelRightScreen(4+1, 0, LED_BLUE);
  } else {
    // Draw a defeat pattern (e.g., "L" for lose)
    drawPixelRightScreen(2, 0, LED_PINK);
    drawPixelRightScreen(2, 1, LED_PINK);
    drawPixelRightScreen(2, 2, LED_PINK);
    drawPixelRightScreen(3, 0, LED_PINK);
    drawPixelRightScreen(4, 0, LED_PINK);
  }
  matrix.writeDisplay();
    // delay(500);
    // matrix.clear();
    // matrix.writeDisplay();
    // delay(300);
  // }
  if (iWon) {
    sendGameOver();
  }
  // redrawScreen();
}

void setup() {
  Serial.begin(115200);
  // while (!Serial) {; /* wait until Serial is ready */}
  delay(100);
  Serial.print("Initializing radio with address #");
  Serial.println(MY_ADDRESS);
  
  // Initialize random seed from analog noise
  randomSeed(analogRead(BTN_DOWN));
  
  // Reset the radio
  pinMode(RFM69_RST, OUTPUT);
  digitalWrite(RFM69_RST, LOW);
  delay(10);
  digitalWrite(RFM69_RST, HIGH);
  delay(10);
  digitalWrite(RFM69_RST, LOW);
  delay(10);
  // Initialize the radio
  if (!manager.init()) {
    Serial.println("RFM69 init failed");
    while (1);
  }
  rf69.setFrequency(915.0);
  rf69.setTxPower(18, true);
  
  manager.setRetries(MAX_RETRIES);
  manager.setTimeout(200);
  
  matrix.begin(0x70);  // default I2C address
  // matrix.setBrightness(2);
  matrix.setRotation(2);
  matrix.clear();
  
  // Initialize button pins as inputs with pull-up resistors
  pinMode(BTN_UP, INPUT);
  pinMode(BTN_DOWN, INPUT);
  pinMode(BTN_LEFT, INPUT);
  pinMode(BTN_RIGHT, INPUT);
  pinMode(BTN_DRAW, INPUT);

  strip.begin();
  strip.setBrightness(50); // Set moderate brightness (0-255)
  strip.show();

  gameOver = false;
  myShipBoardInit();
  
  if (MY_ADDRESS == 2) {
    setMyTurnStatus(true);
  } else {
    setMyTurnStatus(false);
  }
  redrawScreen();

  matrix.setTextWrap(false);  // we dont want text to wrap so it scrolls nicely
  matrix.setTextSize(1);


}


void setIdle(bool status)
{
  isIdle = status;
  if (isIdle){matrix.setRotation(0);Serial.println("switched to idle");}
  else{matrix.setRotation(2);previousMillisIdle = millis();Serial.println("switched to game screen");redrawScreen();}
}

void loop() {
  unsigned long currentMillis = millis();
  if (currentMillis - previousMillisIdle > intervalIdle) {
    previousMillisIdle = currentMillis;
    if (!isIdle){setIdle(true);}
  }
  receiveMessage();
  update();
  // Small delay to prevent excessive CPU usage
  delay(10);
}